/*  ---------------------------------------------------------------  */
/*           The HMM-Based Speech Synthesis System (HTS)             */
/*                       HTS Working Group                           */
/*                                                                   */
/*                  Department of Computer Science                   */
/*                  Nagoya Institute of Technology                   */
/*                               and                                 */
/*   Interdisciplinary Graduate School of Science and Engineering    */
/*                  Tokyo Institute of Technology                    */
/*                                                                   */
/*                     Copyright (c) 2001-2007                       */
/*                       All Rights Reserved.                        */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The source code must retain the above copyright notice,     */
/*       this list of conditions and the following disclaimer.       */
/*                                                                   */
/*    2. Any modifications to the source code must be clearly        */
/*       marked as such.                                             */
/*                                                                   */
/*    3. Redistributions in binary form must reproduce the above     */
/*       copyright notice, this list of conditions and the           */
/*       following disclaimer in the documentation and/or other      */
/*       materials provided with the distribution.  Otherwise, one   */
/*       must contact the HTS working group.                         */
/*                                                                   */
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF TECHNOLOGY,   */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF         */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */

/* $Id: HTS_engine.h,v 1.2 2007/09/28 05:39:58 zen Exp $ */

#include <stdio.h>

typedef int HTS_Boolean;

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif                          /* !TRUE */

typedef enum { DUR, LF0, MCP } HTS_Mtype;

#define HTS_NUMMTYPE 3

/* HTS_Model: structure for individual HMM */
typedef struct _HTS_Model {
   char *name;                  /* the name of this HMM */
   char *lab;                   /* label of this HMM */
   int durpdf;                  /* duration pdf index for this HMM */
   int *lf0pdf;                 /* mel-cepstrum pdf indexes for each state of this HMM */
   int *mcppdf;                 /* log f0 pdf indexes for each state of this HMM */
   int *dur;                    /* duration for each state of this HMM */
   int totaldur;                /* total duration of this HMM */
   double **lf0mean;            /* mean vector of log f0 pdfs for each state of this HMM */
   double **lf0variance;        /* variance (diag) elements of log f0 for each state of this HMM */
   double **mcpmean;            /* mean vector of mel-cepstrum pdfs for each state of this HMM */
   double **mcpvariance;        /* variance (diag) elements of mel-cepstrum for each state of this HMM */
   HTS_Boolean *voiced;         /* voiced/unvoiced decision for each state of this HMM */
   struct _HTS_Model *next;     /* pointer to next HMM */

   /* for gtalk */
   HTS_Boolean bool_rate;
   HTS_Boolean bool_dur;
   HTS_Boolean bool_f0_level;
   HTS_Boolean bool_f0_range;
   HTS_Boolean bool_volume;
   HTS_Boolean bool_alpha;
   double lab_rate;
   int lab_dur;
   double lab_f0_level;
   double lab_f0_range;
   double lab_volume;
   double lab_alpha;

} HTS_Model;

/* HTS_UttMode: structure for utterance HMM */
typedef struct _HTS_UttModel {
   HTS_Model *mhead;
   HTS_Model *mtail;
   int nModel;                  /* # of models for current utterance */
   int nState;                  /* # of HMM states for current utterance */
   int totalframe;              /* # of frames for current utterance */

} HTS_UttModel;

/* HTS_ModelSet: structure for HMM set */
typedef struct _HTS_ModelSet {
   int nstate;                  /* # of HMM states for individual HMM */
   int lf0stream;               /* # of stream for log f0 modeling */
   int mcpvsize;                /* vector size for mel-cepstrum modeling */
   int *nlf0pdf;                /* # of pdfs for each state position (log F0) */
   int *nmcppdf;                /* # of pdfs for each state position (mel-cepstrum) */
   int ndurpdf;                 /* # of pdfs for duration */
   double **durpdf;             /* pdfs for duration */
   double ***mcppdf;            /* pdfs for mel-cepstrum */
   double ****lf0pdf;           /* pdfs for log F0 */
   double weight_interp;        /* weight for interpolation */
} HTS_ModelSet;

/* HTS_Pattern: structure for pattern */
typedef struct _HTS_Pattern {
   char *pat;                   /* pattern */
   struct _HTS_Pattern *next;   /* link to next pattern */
} HTS_Pattern;

/* HTS_Question: structure for questions */
typedef struct _HTS_Question {
   char *qName;                 /* name of this question */
   HTS_Pattern *phead;          /* link to head of pattern list */
   HTS_Pattern *ptail;          /* link to tail of pattern list */
   struct _HTS_Question *next;  /* link to next question */
} HTS_Question;

/* HTS_Node: structure for node of decision tree */
typedef struct _HTS_Node {
   int idx;                     /* index of this node */
   int pdf;                     /* index of pdf for this node  ( leaf node only ) */
   struct _HTS_Node *yes;       /* link to child node (yes) */
   struct _HTS_Node *no;        /* link to child node (no) */
   struct _HTS_Node *next;      /* link to next node */
   HTS_Question *quest;         /* question applied at this node */
} HTS_Node;

/* HTS_Tree: structure for each decision tree */
typedef struct _HTS_Tree {
   int state;                   /* state position of this tree */
   HTS_Pattern *phead;          /* link to head of pattern list for this tree */
   HTS_Pattern *ptail;          /* link to tail of pattern list for this tree */
   struct _HTS_Tree *next;      /* link to next tree */
   HTS_Node *root;              /* root node of this decision tree */
   HTS_Node *leaf;              /* leaf node of this decision tree */
} HTS_Tree;

/* HTS_TreeSet: structure for decision tree set */
typedef struct _HTS_TreeSet {
   HTS_Question *qhead[HTS_NUMMTYPE];   /* question lists for mel-cepstrum, log F0, and duration */
   HTS_Question *qtail[HTS_NUMMTYPE];
   HTS_Tree *thead[HTS_NUMMTYPE];       /* tree lists for mel-cepstrum, log F0, and duration */
   HTS_Tree *ttail[HTS_NUMMTYPE];
   int nTrees[HTS_NUMMTYPE];    /* # of trees for mel-cepstrum, log F0, and duration */
} HTS_TreeSet;

/* HTS_DWin: structure for regression window */
typedef struct _HTS_DWin {
   int num;                     /* number of static + deltas */
   int **width;                 /* width [0..num-1][0(left) 1(right)] */
   double **coef;               /* coefficient [0..num-1][length[0]..length[1]] */
   int maxw[2];                 /* max width [0(left) 1(right)] */
   int max_L;                   /* max {maxw[0], maxw[1]} */
} HTS_DWin;

/* HTS_SMatrices: structure for matrices to generate sequence of speech parameter vectors */
typedef struct _HTS_SMatrices {
   double **mseq;               /* sequence of mean vector */
   double **ivseq;              /* sequence of invarsed variance vector */
   double *g;                   /* for forward substitution */
   double **WUW;                /* W' U^-1 W  */
   double *WUM;                 /* W' U^-1 mu */
} HTS_SMatrices;

/* HTS_PStream: structure for parameter generation setting */
typedef struct _HTS_PStream {
   int vSize;                   /* vector size of observation vector (include static and dynamic features) */
   int order;                   /* vector size of static features */
   int T;                       /* length */
   int width;                   /* width of dynamic window */
   HTS_DWin dw;                 /* dynamic window */
   double **par;                /* output parameter vector */
   HTS_SMatrices sm;            /* matrices for parameter generation */
   HTS_Boolean *voiced;         /* voiced/unvoiced decision */
} HTS_PStream;

/* HTS_globalP: structure for global parameter */
typedef struct _HTS_globalP {
   int rate;
   int fperiod;
   double rho;
   double alpha;
   double beta;
   double f0_std;
   double f0_mean;
   double uv;
   double length;
   HTS_Boolean algnst;
   HTS_Boolean algnph;
   int totaldur;
   int totalframe;
   int nsample;

   /* for audio device */
   int buff_size;

   HTS_Boolean stored_raw_data;
   short *raw_data;

} HTS_globalP;

#if !defined(AUDIO_PLAY_WIN32) && !defined(AUDIO_PLAY_WINCE)
#define AUDIO_PLAY_NONE
#endif                          /* !AUDIO_PLAY_WIN32 && !AUDIO_PLAY_WINCE */

/* define HTS_AudioSet for windows & windows mobile */
#if defined (AUDIO_PLAY_WIN32) || defined(AUDIO_PLAY_WINCE)
#include<afxwin.h>
#include<mmsystem.h>
#ifdef AUDIO_PLAY_WIN32
#pragma comment(lib,"winmm.lib")
#endif                          /* AUDIO_PLAY_WIN32 */
typedef struct _HTS_AudioSet {
   HWAVEOUT hwaveout;
   WAVEFORMATEX waveformatex;
   short *buff;
   int buff_size;
   int which_buff;
   HTS_Boolean now_buff_1;
   HTS_Boolean now_buff_2;
   WAVEHDR buff_1;
   WAVEHDR buff_2;
   int max_buff_size;
} HTS_AudioSet;
#endif                          /* AUDIO_PLAY_WIN32 || AUDIO_PLAY_WINCE */

/* define HTS_AudioSet for others */
#ifdef AUDIO_PLAY_NONE
typedef struct _HTS_AudioSet {
   int i;                       /* make compiler happy */
} HTS_AudioSet;
#endif                          /* AUDIO_PLAY_NONE */

/* HTS_VocoderSet: structure for setting of vocoder */
typedef struct _HTS_VocoderSet {
   int fprd;
   int iprd;
   int seed;
   int pd;
   unsigned long next;
   HTS_Boolean gauss;
   double p1;
   double pc;
   double pade[21];
   double *ppade;
   double *c, *cc, *cinc, *d1;
   double rate;
   int sw;
   double r1, r2, s;
   int x;
   HTS_AudioSet *as;
   /* for postfiltering */
   int size;
   double *d;
   double *g;
   double *mc;
   double *cep;
   double *ir;
   int o;
   int irleng;
} HTS_VocoderSet;

typedef struct _HTS_Engine {
   HTS_ModelSet *ms;
   HTS_TreeSet *ts;
   HTS_PStream lf0pst;
   HTS_PStream mcppst;
   HTS_globalP gp;
   int num_interp;
} HTS_Engine;

/* HTS_Engine_load_fn: setup by file names */
void HTS_Engine_load_fn(HTS_Engine * engine,
                        char **fn_ms_lf0, char **fn_ms_mcp,
                        char **fn_ms_dur, char **fn_ts_lf0,
                        char **fn_ts_mcp, char **fn_ts_dur,
                        int num_ws_lf0, char **fn_ws_lf0,
                        int num_ws_mcp, char **fn_ws_mcp,
                        double *rate_interp, int num_interp);

/* HTS_Engine_load_fp: setup by file pointers */
void HTS_Engine_load_fp(HTS_Engine * engine,
                        FILE ** fp_ms_lf0, FILE ** fp_ms_mcp,
                        FILE ** fp_ms_dur, FILE ** fp_ts_lf0,
                        FILE ** fp_ts_mcp, FILE ** fp_ts_dur,
                        int num_ws_lf0, FILE ** fp_ws_lf0, int num_ws_mcp,
                        FILE ** fp_ws_mcp, double *rate_interp, int num_interp);

/* HTS_Engine_initialize: initialize */
void HTS_Engine_initialize(HTS_Engine * engine);

/* HTS_Engine_set_sampling_rate: set sampling rate */
void HTS_Engine_set_sampling_rate(HTS_Engine * engine, int i);

/* HTS_Engine_set_fperiod: set fperiod */
void HTS_Engine_set_fperiod(HTS_Engine * engine, int i);

/* HTS_Engine_set_alpha: set alpha */
void HTS_Engine_set_alpha(HTS_Engine * engine, double f);

/* HTS_Engine_set_beta: set beta */
void HTS_Engine_set_beta(HTS_Engine * engine, double f);

/* HTS_Engine_set_rho: set rho */
void HTS_Engine_set_rho(HTS_Engine * engine, double f);

/* HTS_Engine_set_f0_std: set f0_std */
void HTS_Engine_set_f0_std(HTS_Engine * engine, double f);

/* HTS_Engine_set_f0_mean: set f0_mean */
void HTS_Engine_set_f0_mean(HTS_Engine * engine, double f);

/* HTS_Engine_setuv: set uv*/
void HTS_Engine_set_uv(HTS_Engine * engine, double f);

/* HTS_Engine_set_length: set length */
void HTS_Engine_set_length(HTS_Engine * engine, double f);

/* HTS_Engine_set_algnst: set algnst */
void HTS_Engine_set_algnst(HTS_Engine * engine, HTS_Boolean i);

/* HTS_Engine_set_algnph: set algnph */
void HTS_Engine_set_algnph(HTS_Engine * engine, HTS_Boolean i);

/* HTS_Engine_set_buff_size: set audio buffer size */
void HTS_Engine_set_buff_size(HTS_Engine * engine, int i);

/* HTS_Engine_get_sampling_rate: get sampling rate */
int HTS_Engine_get_sampling_rate(HTS_Engine * engine);

/* HTS_Engine_get_total_dur: get total dur */
int HTS_Engine_get_total_dur(HTS_Engine * engine);

/* HTS_Engine_get_total_frame: get total frame */
int HTS_Engine_get_total_frame(HTS_Engine * engine);

/* HTS_Engine_get_nsample: get nsample */
int HTS_Engine_get_nsample(HTS_Engine * engine);

/* HTS_Engine_clear: free memory */
void HTS_Engine_clear(HTS_Engine * engine);

/* HTS_Engine_refresh: free model list */
void HTS_Engine_refresh(HTS_Engine * engine, HTS_Model * mhead);

/* HTS_Engine_process: parse label, determine state duration,
   generate sequence of speech parameter vector, and synthesize waveform */
void HTS_Engine_process(HTS_Engine * engine, HTS_Model * mhead,
                        FILE * wavfp, FILE * rawfp, FILE * lf0fp,
                        FILE * mcpfp, FILE * durfp, FILE * tracefp,
                        double *fr_f0, double *fr_power);

/* HTS_Model_load_from_labfp: create model list from file pointer */
void HTS_Model_load_from_labfp(HTS_Model * mhead, FILE * labfp);

/* HTS_Model_load_from_labfn: create model list from file name */
void HTS_Model_load_from_labfn(HTS_Model * mhead, char *fn);

/* HTS_Model_load_from_string: create model list from string */
void HTS_Model_load_from_string(HTS_Model * mhead, char *labdata);

/* HTS_Model_load_from_string_list: create model list from string list */
void HTS_Model_load_from_string_list(HTS_Model * mhead, char **labdata,
                                     int size);

/* HTS_Model_set_rate: set model rate */
void HTS_Model_set_rate(HTS_Model * m, int index, double f);

/* HTS_Model_set_dur: set model duration */
void HTS_Model_set_dur(HTS_Model * m, int index, int i);

/* HTS_Model_set_f0_level: set model f0 level */
void HTS_Model_set_f0_level(HTS_Model * m, int index, double f);

/* HTS_Model_set_f0_range: set model f0 range */
void HTS_Model_set_f0_range(HTS_Model * m, int index, double f);

/* HTS_Model_set_volume: set model volume */
void HTS_Model_set_volume(HTS_Model * m, int index, double f);

/* HTS_Model_set_alpha: set model alpha */
void HTS_Model_set_alpha(HTS_Model * m, int index, double f);

/* HTS_Engine_get_pros_len: get pros length */
int HTS_Engine_get_pros_len(HTS_Engine * engine);

/* HTS_Engine_get_pros: get pros */
HTS_Boolean HTS_Engine_get_pros(HTS_Engine * engine, int len,
                                double *f0_data, double *power_data);

void HTS_Engine_set_stored_raw_data(HTS_Engine * engine, HTS_Boolean i);
short *HTS_Engine_get_stored_raw_data(HTS_Engine * engine);

/* Function prototypes for vocoding */
void HTS_VocoderSet_initialize(HTS_VocoderSet * vs, const int m,
                               const int rate, const int fperiod,
                               int buff_size);
void HTS_VocoderSet_synthesize(HTS_VocoderSet * vs, const int m,
                               double p, double *mc, double alpha,
                               double beta, FILE * wavfp, FILE * rawfp,
                               short *rawdata);
void HTS_VocoderSet_clear(HTS_VocoderSet * vs);
