/*  ---------------------------------------------------------------  */
/*           The HMM-Based Speech Synthesis System (HTS)             */
/*                       HTS Working Group                           */
/*                                                                   */
/*                  Department of Computer Science                   */
/*                  Nagoya Institute of Technology                   */
/*                               and                                 */
/*   Interdisciplinary Graduate School of Science and Engineering    */
/*                  Tokyo Institute of Technology                    */
/*                                                                   */
/*                     Copyright (c) 2001-2007                       */
/*                       All Rights Reserved.                        */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The source code must retain the above copyright notice,     */
/*       this list of conditions and the following disclaimer.       */
/*                                                                   */
/*    2. Any modifications to the source code must be clearly        */
/*       marked as such.                                             */
/*                                                                   */
/*    3. Redistributions in binary form must reproduce the above     */
/*       copyright notice, this list of conditions and the           */
/*       following disclaimer in the documentation and/or other      */
/*       materials provided with the distribution.  Otherwise, one   */
/*       must contact the HTS working group.                         */
/*                                                                   */
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF TECHNOLOGY,   */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF         */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */

/* $Id: HTS_hidden.h,v 1.2 2007/09/28 05:40:09 zen Exp $ */

#include "HTS_engine.h"

/*--------------------------- audio ---------------------------------*/

void HTS_AudioSet_open(HTS_AudioSet * as, int sampling_rate, int max_buff_size);
void HTS_AudioSet_write(HTS_AudioSet * as, short data);
void HTS_AudioSet_close(HTS_AudioSet * as);

/*--------------------------- misc ----------------------------------*/

#define HTS_MAXBUFLEN 1024

/* ----- routines for file input/output -----*/
void HTS_error(const int, char *, ...);
FILE *HTS_get_fp(const char *, const char *);
void HTS_get_token(FILE *, char *);
void HTS_get_token_from_str(char *, int *, char *);
HTS_Boolean HTS_get_line(FILE *, char *);
HTS_Boolean HTS_get_line_from_str(char *, int *, char *);

/* ----- routines for memory allocation/free -----*/
char *HTS_calloc(const size_t, const size_t);
void HTS_free(void *);
char *HTS_strdup(const char *);
double *HTS_alloc_vector(const int);
double **HTS_alloc_matrix(const int, const int);
void HTS_free_vector(double *);
void HTS_free_matrix(double **, const int);

/* ----- Routines for reading from binary file ----- */
int HTS_fread_big_endian(void *, const int, const int, FILE *);
/* ----- Routines for writing from binary file ----- */
int HTS_fwrite_little_endian(void *, const int, const int, FILE *);

/* ----- Routines for reading from ascii file ----- */
double HTS_read_double(FILE * fp);
int HTS_read_int(FILE * fp);

/*--------------------------- mlpg ----------------------------------*/

#define INFTY   ((double) 1.0e+38)
#define INFTY2  ((double) 1.0e+19)
#define INVINF  ((double) 1.0e-38)
#define INVINF2 ((double) 1.0e-19)
#define LTPI    1.83787706640935        /* log(2*PI) */

#define WLEFT  0
#define WRIGHT 1

/* Function prototype for HMM-based speech synthesis */
void HTS_pdf2speech(FILE * wavfp, FILE * rawfp,
                    FILE * lf0fp, FILE * mcpfp, double *fr_f0,
                    double *fr_power, HTS_PStream * lf0pst,
                    HTS_PStream * mcppst, HTS_globalP * gp,
                    HTS_ModelSet * ms, HTS_UttModel * um);
/* Initialise dynamic window */
void HTS_PStream_load(HTS_PStream * lf0pst, HTS_PStream * mcppst,
                      FILE ** fp_ws_lf0, FILE ** fp_ws_mcp,
                      int num_ws_lf0, int num_ws_mcp, int mcpvsize);
/* free regression window */
void HTS_PStream_clear(HTS_PStream * lf0pst, HTS_PStream * mcppst);
void HTS_PStream_refresh(HTS_PStream * lf0pst, HTS_PStream * mcppst);

/*-------------------------- model ----------------------------------*/

/* Function prototypes for tree handling */
void HTS_HTS_ModelSet_load(HTS_ModelSet *, FILE *, FILE *, FILE *);
void HTS_HTS_ModelSet_clear(HTS_ModelSet *);
void HTS_Model_determine_lf0(HTS_Model * m, HTS_ModelSet * ms,
                             HTS_TreeSet * ts, const int num_interp,
                             double uvthresh);
void HTS_Model_determine_mcp(HTS_Model * m, HTS_ModelSet * ms,
                             HTS_TreeSet * ts, const int num_interp);
void HTS_Model_determine_dur(HTS_Model * m, HTS_ModelSet * ms,
                             double *durmean, double *durvar,
                             double rho, double *diffdur);
void HTS_Model_estimate_dur(HTS_Model * m, HTS_ModelSet * ms,
                            HTS_TreeSet * ts, const int num_interp,
                            double *durmean, double *durvar);

/*---------------------------- tree ---------------------------------*/

/* Function prototypes for tree handling */
void HTS_TreeSet_load(HTS_TreeSet * ts, FILE * fp, const HTS_Mtype type);
void HTS_TreeSet_clear(HTS_TreeSet * ts, const HTS_Mtype type);
int HTS_Node_search(HTS_Node * node, char *str);

/*-------------------------- vocoder --------------------------------*/

#define RANDMAX 32767

#define   IPERIOD    1
#define   SEED       1
#define   B0         0x00000001
#define   B28        0x10000000
#define   B31        0x80000000
#define   B31_       0x7fffffff
#define   Z          0x00000000

#ifdef HTS_EMBEDDED
#define   GAUSS      FALSE
#define   PADEORDER  4          /* pade order (for MLSA filter) */
#define   IRLENG     64         /* length of impulse response */
#else
#define   GAUSS      TRUE
#define   PADEORDER  5
#define   IRLENG     96
#endif                          /* HTS_EMBEDDED */
