/*  ---------------------------------------------------------------  */
/*           The HMM-Based Speech Synthesis System (HTS)             */
/*                       HTS Working Group                           */
/*                                                                   */
/*                  Department of Computer Science                   */
/*                  Nagoya Institute of Technology                   */
/*                               and                                 */
/*   Interdisciplinary Graduate School of Science and Engineering    */
/*                  Tokyo Institute of Technology                    */
/*                                                                   */
/*                     Copyright (c) 2001-2007                       */
/*                       All Rights Reserved.                        */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The source code must retain the above copyright notice,     */
/*       this list of conditions and the following disclaimer.       */
/*                                                                   */
/*    2. Any modifications to the source code must be clearly        */
/*       marked as such.                                             */
/*                                                                   */
/*    3. Redistributions in binary form must reproduce the above     */
/*       copyright notice, this list of conditions and the           */
/*       following disclaimer in the documentation and/or other      */
/*       materials provided with the distribution.  Otherwise, one   */
/*       must contact the HTS working group.                         */
/*                                                                   */
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF TECHNOLOGY,   */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF         */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */

/* $Id: HTS_misc.c,v 1.2 2007/09/28 05:40:09 zen Exp $ */

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

/* hts_engine libraries */
#include "HTS_hidden.h"

#ifdef FESTIVAL
#include "EST_walloc.h"
#endif                          /* FESTIVAL */

/* ----- Routines for file input/output ----- */
/* HTS_error: output error message */
void HTS_error(const int error, char *message, ...)
{
   va_list arg;

   fflush(stdout);
   fflush(stderr);

   if (error > 0)
      fprintf(stderr, "\nError: ");
   else
      fprintf(stderr, "\nWarning: ");

   va_start(arg, message);
   vfprintf(stderr, message, arg);
   va_end(arg);

   fflush(stderr);

   if (error > 0)
      exit(error);
}

/* HTS_get_fp: wrapper for fopen */
FILE *HTS_get_fp(const char *name, const char *opt)
{
   FILE *fp = fopen(name, opt);

   if (fp == NULL)
      HTS_error(2, "HTS_get_fp: Cannot open %s.\n", name);

   return (fp);
}

/* HTS_get_token: parser */
void HTS_get_token(FILE * fp, char *buff)
{
   char c;
   int i;
   HTS_Boolean squote = FALSE, dquote = FALSE;

   c = fgetc(fp);

   while (c == ' ' || c == '\n')
      c = fgetc(fp);

   if (c == '\'') {             /* single quote case */
      c = fgetc(fp);
      squote = TRUE;
   }

   if (c == '\"') {             /*double quote case */
      c = fgetc(fp);
      dquote = TRUE;
   }

   if (c == ',') {              /*special character ',' */
      strcpy(buff, ",");
      return;
   }

   i = 0;
   while (1) {
      buff[i++] = c;
      c = fgetc(fp);
      if (squote && c == '\'')
         break;
      if (dquote && c == '\"')
         break;
      if (!squote && !dquote) {
         if (c == ' ')
            break;
         if (c == '\n')
            break;
         if (feof(fp))
            break;
      }
   }

   buff[i] = '\0';
}


/* HTS_get_token_from_str: parser */
void HTS_get_token_from_str(char *data, int *dp, char *buff)
{
   char c;
   int i;
   HTS_Boolean squote = FALSE, dquote = FALSE;

   c = data[(*dp)++];

   while (isspace(c)) {
      c = data[(*dp)++];
   }

   if (c == '\'') {             /* single quote case */
      c = data[(*dp)++];
      squote = TRUE;
   }

   if (c == '\"') {             /*double quote case */
      c = data[(*dp)++];
      dquote = TRUE;
   }

   if (c == ',') {              /*special character ',' */
      strcpy(buff, ",");
      return;
   }

   i = 0;
   while (1) {
      buff[i++] = c;
      c = data[(*dp)++];
      if (squote && c == '\'')
         break;
      if (dquote && c == '\"')
         break;
      if (!(squote || dquote || isgraph(c)))
         break;
   }

   buff[i] = '\0';
}

/* HTS_get_line: get one line */
HTS_Boolean HTS_get_line(FILE * fp, char *buff)
{
   int i;
   if (!fgets(buff, HTS_MAXBUFLEN, fp)) {
      return FALSE;
   }
   i = strlen(buff);
   if (buff[i - 1] == '\n') {
      buff[i - 1] = 0;
   }
   return TRUE;
}

/* HTS_get_line_from_str: get one line */
HTS_Boolean HTS_get_line_from_str(char *data, int *dp, char *buff)
{
   char c;
   int i;

   c = data[(*dp)++];

   /* skip spaces */
   while (isspace(c)) {
      c = data[(*dp)++];
   }

   i = 0;
   while (1) {
      if (c == '\n') {
         if (i > 0) {
            break;
         }
      } else if (c == '\0') {
         if (i > 0) {
            (*dp)--;
            break;
         } else {
            return FALSE;
         }
      } else {
         buff[i++] = c;
      }
      c = data[(*dp)++];
   }
   buff[i] = 0;

   return TRUE;
}

/* ----- Routines for memory allocation/free ----- */
/* HTS_calloc: wrapper for calloc */
char *HTS_calloc(const size_t num, const size_t size)
{
#ifdef FESTIVAL
   char *mem = (char *) safe_wcalloc(num * size);
#else
   char *mem = (char *) calloc(num, size);
#endif                          /* FESTIVAL */

   if (mem == NULL)
      HTS_error(1, "HTS_calloc: Cannot allocate memory.\n");

   return (mem);
}

/* HTS_Free: wrapper for free */
void HTS_free(void *ptr)
{
#ifdef FESTIVAL
   wfree(ptr);
#else
   free(ptr);
#endif                          /* FESTIVAL */
}

/* HTS_strdup: wrapper for strdup */
char *HTS_strdup(const char *in)
{
#ifdef FESTIVAL
   return (wstrdup(in));
#else
   char *tmp = (char *) HTS_calloc(strlen(in) + 1, sizeof(char));
   strcpy(tmp, in);
   return tmp;
#endif                          /* FESTIVAL */
}

/* HTS_AllocVector: allocate vector */
double *HTS_alloc_vector(const int x)
{
   double *ptr = (double *) HTS_calloc(x, sizeof(double));

   ptr--;

   return (ptr);
}

/* HTS_alloc_matrix: allocate matrix */
double **HTS_alloc_matrix(const int x, const int y)
{
   int i;
   double **ptr = (double **) HTS_calloc(x, sizeof(double *));

   ptr--;

   for (i = 1; i <= x; i++)
      ptr[i] = HTS_alloc_vector(y);

   return (ptr);
}

/* HTS_free_vector: free vector */
void HTS_free_vector(double *ptr)
{
   ptr++;
   HTS_free((void *) ptr);
}

/* HTS_free_matrix: free matrix */
void HTS_free_matrix(double **ptr, const int x)
{
   int i;

   for (i = x; i > 0; i--)
      HTS_free_vector(ptr[i]);

   ptr++;

   HTS_free((void *) ptr);
}

/* ----- Routines for reading from binary file ----- */
/* HTS_byte_swap: byte swap */
static int HTS_byte_swap(void *p, const int size, const int blocks)
{
   char *q, tmp;
   int i, j;

   q = (char *) p;

   for (i = 0; i < blocks; i++) {
      for (j = 0; j < (size / 2); j++) {
         tmp = *(q + j);
         *(q + j) = *(q + (size - 1 - j));
         *(q + (size - 1 - j)) = tmp;
      }
      q += size;
   }

   return i;
}

#if !defined(WORDS_BIGENDIAN) && !defined(WORDS_LITTLEENDIAN)
#define WORDS_LITTLEENDIAN
#endif                          /* !WORDS_BIGENDIAN && !WORDS_LITTLEENDIAN */
#if defined(WORDS_BIGENDIAN) && defined(WORDS_LITTLEENDIAN)
#undef WORDS_BIGENDIAN
#endif                          /* WORDS_BIGENDIAN && WORDS_LITTLEENDIAN */

/* HTS_fread: fread with byteswap */
int HTS_fread_big_endian(void *p, const int size, const int num, FILE * fp)
{
   const int block = fread(p, size, num, fp);

#ifdef WORDS_LITTLEENDIAN
   HTS_byte_swap(p, size, block);
#endif                          /* WORDS_LITTLEENDIAN */

   return block;
}

/* HTS_fwrite: fwrite with byteswap */
int HTS_fwrite_little_endian(void *p, const int size, const int num, FILE * fp)
{
   const int block = num * size;

#ifdef WORDS_BIGENDIAN
   HTS_byte_swap(p, size, block);
#endif                          /* WORDS_BIGENDIAN */
   fwrite(p, size, num, fp);

   return block;
}

/* HTS_read_double: read one double from ascii file */
double HTS_read_double(FILE * fp)
{
   double d;

   if (fscanf(fp, "%lf", &d) != 1)
      return 0.0;
   return d;
}

/* HTS_read_int: read one integer from ascii file */
int HTS_read_int(FILE * fp)
{
   int i;

   if (fscanf(fp, "%d", &i) != 1)
      return 0;
   return i;
}
