/*  ---------------------------------------------------------------  */
/*           The HMM-Based Speech Synthesis System (HTS)             */
/*                       HTS Working Group                           */
/*                                                                   */
/*                  Department of Computer Science                   */
/*                  Nagoya Institute of Technology                   */
/*                               and                                 */
/*   Interdisciplinary Graduate School of Science and Engineering    */
/*                  Tokyo Institute of Technology                    */
/*                                                                   */
/*                     Copyright (c) 2001-2007                       */
/*                       All Rights Reserved.                        */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The source code must retain the above copyright notice,     */
/*       this list of conditions and the following disclaimer.       */
/*                                                                   */
/*    2. Any modifications to the source code must be clearly        */
/*       marked as such.                                             */
/*                                                                   */
/*    3. Redistributions in binary form must reproduce the above     */
/*       copyright notice, this list of conditions and the           */
/*       following disclaimer in the documentation and/or other      */
/*       materials provided with the distribution.  Otherwise, one   */
/*       must contact the HTS working group.                         */
/*                                                                   */
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF TECHNOLOGY,   */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF         */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */

/* $Id: HTS_mlpg.c,v 1.2 2007/09/28 05:40:09 zen Exp $ */

#include <math.h>

/* hts_engine libraries */
#include "HTS_hidden.h"

static double HTS_finv(const double x)
{
   if (x >= INFTY2)
      return 0.0;
   if (x <= -INFTY2)
      return 0.0;
   if (x <= INVINF2 && x >= 0)
      return INFTY;
   if (x >= -INVINF2 && x < 0)
      return -INFTY;

   return (1.0 / x);
}

/* HTS_PStream_calc_wuw_and_wum: calcurate W'U^{-1}W and W'U^{-1}M */
static void HTS_PStream_calc_wuw_and_wum(HTS_PStream * pst, const int m)
{
   int t, i, j, k;
   double WU;

   for (t = 1; t <= pst->T; t++) {
      /* initialize */
      pst->sm.WUM[t] = 0.0;
      for (i = 1; i <= pst->width; i++)
         pst->sm.WUW[t][i] = 0.0;

      /* calc WUW & WUM */
      for (i = 0; i < pst->dw.num; i++)
         for (j = pst->dw.width[i][0]; j <= pst->dw.width[i][1]; j++)
            if ((t + j > 0) && (t + j <= pst->T)
                && (pst->dw.coef[i][-j] != 0.0)) {
               WU = pst->dw.coef[i][-j] *
                   pst->sm.ivseq[t + j][i * pst->order + m];
               pst->sm.WUM[t] += WU * pst->sm.mseq[t + j][i * pst->order + m];
               for (k = 0; (k < pst->width) && (t + k <= pst->T); k++)
                  if ((k - j <= pst->dw.width[i][1])
                      && (pst->dw.coef[i][k - j] != 0.0))
                     pst->sm.WUW[t][k + 1] += WU * pst->dw.coef[i][k - j];
            }
   }
}


/* HTS_PStream_ldl_factorization: Factorize W'*U^{-1}*W to L*D*L' (L: lower triangular, D: diagonal) */
static void HTS_PStream_ldl_factorization(HTS_PStream * pst)
{
   int t, i, j;

   for (t = 1; t <= pst->T; t++) {
      for (i = 1; (i < pst->width) && (t - i > 0); i++)
         pst->sm.WUW[t][1] -=
             pst->sm.WUW[t - i][i + 1] *
             pst->sm.WUW[t - i][i + 1] * pst->sm.WUW[t - i][1];

      for (i = 2; i <= pst->width; i++) {
         for (j = 1; (i + j <= pst->width) && (t - j > 0); j++)
            pst->sm.WUW[t][i] -=
                pst->sm.WUW[t - j][j + 1] *
                pst->sm.WUW[t - j][i + j] * pst->sm.WUW[t - j][1];
         pst->sm.WUW[t][i] /= pst->sm.WUW[t][1];
      }
   }
}

/* HTS_PStream_forward_substitution */
static void HTS_PStream_forward_substitution(HTS_PStream * pst)
{
   int t, i;

   for (t = 1; t <= pst->T; t++) {
      pst->sm.g[t] = pst->sm.WUM[t];
      for (i = 1; (i < pst->width) && (t - i > 0); i++)
         pst->sm.g[t] -= pst->sm.WUW[t - i][i + 1] * pst->sm.g[t - i];
   }
}

/* HTS_PStream_backward_substitution */
static void HTS_PStream_backward_substitution(HTS_PStream * pst, const int m)
{
   int t, i;

   for (t = pst->T; t > 0; t--) {
      pst->par[t][m] = pst->sm.g[t] / pst->sm.WUW[t][1];
      for (i = 1; (i < pst->width) && (t + i <= pst->T); i++)
         pst->par[t][m] -= pst->sm.WUW[t][i + 1] * pst->par[t + i][m];
   }
}

/* HTS_PStream_mlpg: generate sequence of speech parameter vector maximizing its output probability for given pdf sequence */
static void HTS_PStream_mlpg(HTS_PStream * pst)
{
   int m;
   const int M = pst->order;

   for (m = 1; m <= M; m++) {
      HTS_PStream_calc_wuw_and_wum(pst, m);
      HTS_PStream_ldl_factorization(pst);       /* LDL factorization */
      HTS_PStream_forward_substitution(pst);    /* forward substitution   */
      HTS_PStream_backward_substitution(pst, m);        /* backward substitution  */
   }
}

/* HTS_PStream_load_dwin: Initialise dynamic window */
static void HTS_PStream_load_dwin(HTS_PStream * pst, FILE ** fps)
{
   int i, j;
   int fsize, leng;
   FILE *fp;

   /* memory allocation */
   pst->dw.width = (int **) HTS_calloc(pst->dw.num, sizeof(int *));

   for (i = 0; i < pst->dw.num; i++) {
      pst->dw.width[i] = (int *) HTS_calloc(2, sizeof(int));
   }

   pst->dw.coef = (double **) HTS_calloc(pst->dw.num, sizeof(double *));

   /* set delta coefficents */
   for (i = 0; i < pst->dw.num; i++) {
      fp = fps[i];
      fsize = HTS_read_int(fp);
      /* read coefficients */
      pst->dw.coef[i] = (double *) HTS_calloc(fsize, sizeof(double));

      for (j = 0; j < fsize; j++) {
         pst->dw.coef[i][j] = HTS_read_double(fp);
      }

      /* set pointer */
      leng = fsize / 2;
      pst->dw.coef[i] += leng;
      pst->dw.width[i][WLEFT] = -leng;
      pst->dw.width[i][WRIGHT] = leng;

      if (fsize % 2 == 0)
         pst->dw.width[i][WRIGHT]--;
   }

   pst->dw.maxw[WLEFT] = pst->dw.maxw[WRIGHT] = 0;

   for (i = 0; i < pst->dw.num; i++) {
      if (pst->dw.maxw[WLEFT] > pst->dw.width[i][WLEFT])
         pst->dw.maxw[WLEFT] = pst->dw.width[i][WLEFT];
      if (pst->dw.maxw[WRIGHT] < pst->dw.width[i][WRIGHT])
         pst->dw.maxw[WRIGHT] = pst->dw.width[i][WRIGHT];
   }

   /* calcurate max_L to determine size of band matrix */
   if (pst->dw.maxw[WLEFT] >= pst->dw.maxw[WRIGHT])
      pst->dw.max_L = pst->dw.maxw[WLEFT];
   else
      pst->dw.max_L = pst->dw.maxw[WRIGHT];
}

/* HTS_PStream_clear_dwin: free regression window */
static void HTS_PStream_clear_dwin(HTS_PStream * pst)
{
   int i;

   /* free window */
   for (i = pst->dw.num - 1; i >= 0; i--) {
      pst->dw.coef[i] += pst->dw.width[i][WLEFT];
      HTS_free(pst->dw.coef[i]);
   }
   HTS_free(pst->dw.coef);

   for (i = pst->dw.num - 1; i >= 0; i--)
      HTS_free(pst->dw.width[i]);
   HTS_free(pst->dw.width);
}

/* HTS_PStream_load: initialize PStreams */
void HTS_PStream_load(HTS_PStream * lf0pst, HTS_PStream * mcppst,
                      FILE ** fp_ws_lf0, FILE ** fp_ws_mcp,
                      int num_ws_lf0, int num_ws_mcp, int mcpvsize)
{
   int i;

   lf0pst->dw.num = num_ws_lf0;
   mcppst->dw.num = num_ws_mcp;

   /* check */
   for (i = 0; i < num_ws_lf0; i++) {
      if (fp_ws_lf0[i] == NULL)
         HTS_error(1,
                   "HTS_PStream_load: file for lf0 win[%d] is not specified.\n",
                   i);
   }
   for (i = 0; i < num_ws_mcp; i++) {
      if (fp_ws_mcp[i] == NULL)
         HTS_error(1,
                   "HTS_PStream_load: file for mel-cepstrum win[%d] is not specified.\n",
                   i);
   }

   lf0pst->vSize = lf0pst->dw.num;
   lf0pst->order = 1;
   mcppst->vSize = mcpvsize;
   mcppst->order = mcppst->vSize / mcppst->dw.num;

   HTS_PStream_load_dwin(lf0pst, fp_ws_lf0);
   HTS_PStream_load_dwin(mcppst, fp_ws_mcp);
}

/* HTS_PStream_clear: free HTS_PStreams */
void HTS_PStream_clear(HTS_PStream * lf0pst, HTS_PStream * mcppst)
{
   HTS_PStream_clear_dwin(lf0pst);
   HTS_PStream_clear_dwin(mcppst);
}

static void HTS_PStream_set_voiced(HTS_PStream * lf0pst,
                                   HTS_PStream * mcppst, int size)
{
   HTS_Boolean *voiced = (HTS_Boolean *) HTS_calloc(size, sizeof(HTS_Boolean));
   lf0pst->voiced = voiced;
   mcppst->voiced = voiced;
}

/* HTS_PStream_set_length: Initialise HTS_PStream for parameter generation */
static void HTS_PStream_set_length(HTS_PStream * pst, int len)
{
   pst->T = len;
   pst->width = pst->dw.max_L * 2 + 1;  /* band width of R */

   pst->sm.mseq = HTS_alloc_matrix(pst->T, pst->vSize);
   pst->sm.ivseq = HTS_alloc_matrix(pst->T, pst->vSize);
   pst->sm.WUW = HTS_alloc_matrix(pst->T, pst->width);
   pst->par = HTS_alloc_matrix(pst->T, pst->order);
   pst->sm.g = HTS_alloc_vector(pst->T);
   pst->sm.WUM = HTS_alloc_vector(pst->T);
}

void HTS_PStream_refresh(HTS_PStream * lf0pst, HTS_PStream * mcppst)
{
   HTS_free(lf0pst->voiced);

   HTS_free_vector(lf0pst->sm.WUM);
   HTS_free_vector(lf0pst->sm.g);
   HTS_free_matrix(lf0pst->sm.WUW, lf0pst->T);
   HTS_free_matrix(lf0pst->sm.ivseq, lf0pst->T);
   HTS_free_matrix(lf0pst->sm.mseq, lf0pst->T);
   HTS_free_matrix(lf0pst->par, lf0pst->T);
   HTS_free_vector(mcppst->sm.WUM);
   HTS_free_vector(mcppst->sm.g);
   HTS_free_matrix(mcppst->sm.WUW, mcppst->T);
   HTS_free_matrix(mcppst->sm.ivseq, mcppst->T);
   HTS_free_matrix(mcppst->sm.mseq, mcppst->T);
   HTS_free_matrix(mcppst->par, mcppst->T);
}

/* pdf2speech: parameter generation and waveform synthesis */
void HTS_pdf2speech(FILE * wavfp, FILE * rawfp,
                    FILE * lf0fp, FILE * mcpfp, double *fr_f0,
                    double *fr_power, HTS_PStream * lf0pst,
                    HTS_PStream * mcppst, HTS_globalP * gp,
                    HTS_ModelSet * ms, HTS_UttModel * um)
{
   int frame, lf0frame, mcpframe;
   int state, lw, rw, k, n, t, c, d;
   HTS_Model *m;
   HTS_Boolean nobound, *voiced;
   double f0, loglevel, logvolume, pitch_ave, alp;
   float temp;
   double *alpha;
   HTS_VocoderSet vs;
   int raw_data_index = 0;
   short *raw_data_tmp;

   /* initialize MLSA filter */
   HTS_VocoderSet_initialize(&vs, ms->mcpvsize - 1, gp->rate,
                             gp->fperiod, gp->buff_size);

   if (gp->stored_raw_data) {
      gp->raw_data = (short *) HTS_calloc(gp->nsample, sizeof(short));
   }

   lf0frame = 0;
   mcpframe = 0;

   /* voiced/unvoiced decision */
   HTS_PStream_set_voiced(lf0pst, mcppst, um->totalframe + 1);
   voiced = lf0pst->voiced;

   /* alpha */
   alpha = (double *) HTS_calloc(um->totalframe, sizeof(double));
   alpha--;

   if (fr_f0 == NULL) {
      for (m = um->mhead;; m = m->next) {
         for (state = 2; state <= ms->nstate + 1; state++) {
            for (frame = 1; frame <= m->dur[state]; frame++)
               voiced[++mcpframe] = m->voiced[state];
            if (m->voiced[state])
               lf0frame += m->dur[state];
         }
         if (m == um->mtail)
            break;
      }
   } else {
      for (m = um->mhead;; m = m->next) {
         for (state = 2; state <= ms->nstate + 1; state++) {
            for (frame = 1; frame <= m->dur[state]; frame++) {
               mcpframe++;
               if (fr_f0[mcpframe] == 0.0) {
                  voiced[mcpframe] = FALSE;
               } else {
                  voiced[mcpframe] = TRUE;
                  ++lf0frame;
               }
            }
         }
         if (m == um->mtail)
            break;
      }
   }

   /* initialise parameter generation */
   HTS_PStream_set_length(lf0pst, lf0frame);
   HTS_PStream_set_length(mcppst, mcpframe);

   lf0frame = 1;
   mcpframe = 1;

   /* copy pdfs */
   for (m = um->mhead;; m = m->next) {
      for (state = 2; state <= ms->nstate + 1; state++) {
         for (frame = 1; frame <= m->dur[state]; frame++) {
            /* copy pdf for mcp */
            for (k = 0; k < ms->mcpvsize; k++) {
               mcppst->sm.mseq[mcpframe][k + 1] = m->mcpmean[state][k];
               mcppst->sm.ivseq[mcpframe][k + 1] =
                   HTS_finv(m->mcpvariance[state][k]);
            }
            /* copy pdfs for lf0 */
            for (k = 0; k < ms->lf0stream; k++) {
               lw = lf0pst->dw.width[k][WLEFT];
               rw = lf0pst->dw.width[k][WRIGHT];
               nobound = (HTS_Boolean) 1;
               /* check current frame is voiced/unvoiced boundary or not */
               for (n = lw; n <= rw; n++)
                  if (mcpframe + n <= 0 || um->totalframe < mcpframe + n)
                     nobound = (HTS_Boolean) 0;
                  else
                     nobound = (HTS_Boolean)
                         ((int) nobound & (int) voiced[mcpframe + n]);

               /* copy pdfs */
               if (voiced[mcpframe]) {
                  lf0pst->sm.mseq[lf0frame][k + 1] = m->lf0mean[state][k + 1];
                  if (nobound || k == 0)
                     lf0pst->sm.ivseq[lf0frame][k + 1] =
                         HTS_finv(m->lf0variance[state][k + 1]);
                  else          /* the variances for dynamic feature are set to inf on v/uv boundary */
                     lf0pst->sm.ivseq[lf0frame][k + 1] = 0.0;
               }
            }
            if (voiced[mcpframe])
               lf0frame++;
            mcpframe++;
         }
      }
      if (m == um->mtail)
         break;
   }

   /* parameter generation for mcp */
   HTS_PStream_mlpg(mcppst);

   /* parameter generation for lf0 */
   if (lf0frame > 0)
      HTS_PStream_mlpg(lf0pst);

   /*----- modify pitch & power & alpha start -----*/
   c = 1;
   d = 1;
   pitch_ave = 0.0;
   for (m = um->mhead;; m = m->next) {
      if (m->bool_f0_level)
         loglevel = log(m->lab_f0_level);
      else
         loglevel = 0.0;

      if (m->bool_volume)
         logvolume = log(m->lab_volume);
      else
         logvolume = 0.0;

      if (m->bool_alpha)
         alp = m->lab_alpha;
      else
         alp = gp->alpha;

      for (t = 1; t <= m->totaldur; t++) {
         alpha[d] = alp;
         mcppst->par[d][1] += logvolume;
         if (voiced[d]) {
            lf0pst->par[c][1] += loglevel;
            pitch_ave += lf0pst->par[c][1];
            c++;
         }
         d++;
      }
      if (m == um->mtail)
         break;
   }
   pitch_ave /= (c - 1);        /* average */
   c = 1;
   d = 1;
   for (m = um->mhead;; m = m->next) {
      for (t = 1; t <= m->totaldur; t++) {
         if (voiced[d++]) {
            if (m->bool_f0_range)
               lf0pst->par[c][1] +=
                   (lf0pst->par[c][1] - pitch_ave) * m->lab_f0_range;
            c++;
         }
      }
      if (m == um->mtail)
         break;
   }
   /*----- modify pitch & power & alpha end -----*/

   lf0frame = 1;

   /* synthesize speech waveforms */
   for (mcpframe = 1; mcpframe <= mcppst->T; mcpframe++) {
      /* f0 modification */
      if (voiced[mcpframe]) {
         if (fr_f0 == NULL) {
            f0 = gp->f0_std * exp(lf0pst->par[lf0frame++][1]) + gp->f0_mean;
         } else {
            f0 = exp(fr_f0[mcpframe]);
            lf0pst->par[lf0frame++][1] = log((f0 - gp->f0_mean) / gp->f0_std);
         }
      } else {
         f0 = 0.0;
      }
      if (fr_power != NULL) {
         mcppst->par[mcpframe][1] = fr_power[mcpframe];
      }

      if (gp->raw_data == NULL)
         raw_data_tmp = NULL;
      else {
         raw_data_tmp = gp->raw_data + raw_data_index;
         raw_data_index += gp->fperiod;
      }

      /* synthesize waveforms by MLSA filter */
      HTS_VocoderSet_synthesize(&vs, mcppst->order - 1, f0,
                                &mcppst->par[mcpframe][1], alpha[mcpframe],
                                gp->beta, wavfp, rawfp, raw_data_tmp);

      /* output mel-cepstrum sequence */
      if (mcpfp != NULL) {
         for (k = 1; k <= mcppst->order; k++) {
            temp = (float) mcppst->par[mcpframe][k];
            fwrite(&temp, sizeof(float), 1, mcpfp);
         }
      }

      /* output f0 sequence */
      if (lf0fp != NULL) {
         temp = (float) f0;
         fwrite(&temp, sizeof(float), 1, lf0fp);
      }
   }

   HTS_free(++alpha);
   HTS_VocoderSet_clear(&vs);
}
