/*  ---------------------------------------------------------------  */
/*           The HMM-Based Speech Synthesis System (HTS)             */
/*                       HTS Working Group                           */
/*                                                                   */
/*                  Department of Computer Science                   */
/*                  Nagoya Institute of Technology                   */
/*                               and                                 */
/*   Interdisciplinary Graduate School of Science and Engineering    */
/*                  Tokyo Institute of Technology                    */
/*                                                                   */
/*                     Copyright (c) 2001-2007                       */
/*                       All Rights Reserved.                        */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The source code must retain the above copyright notice,     */
/*       this list of conditions and the following disclaimer.       */
/*                                                                   */
/*    2. Any modifications to the source code must be clearly        */
/*       marked as such.                                             */
/*                                                                   */
/*    3. Redistributions in binary form must reproduce the above     */
/*       copyright notice, this list of conditions and the           */
/*       following disclaimer in the documentation and/or other      */
/*       materials provided with the distribution.  Otherwise, one   */
/*       must contact the HTS working group.                         */
/*                                                                   */
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF TECHNOLOGY,   */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF         */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */

/* $Id: HTS_model.c,v 1.2 2007/09/28 05:40:09 zen Exp $ */

/* hts_engine libraries */
#include "HTS_hidden.h"

/* HTS_HTS_ModelSet_load: load models */
void HTS_HTS_ModelSet_load(HTS_ModelSet * ms, FILE * fp_ms_lf0,
                           FILE * fp_ms_mcp, FILE * fp_ms_dur)
{
   int i, j, k, l;
   double vw, uvw;
   float temp;

   /* check */
   if (fp_ms_dur == NULL)
      HTS_error(1,
                "HTS_HTS_ModelSet_load: file for duration pdfs is not specified.\n");
   if (fp_ms_lf0 == NULL)
      HTS_error(1,
                "HTS_HTS_ModelSet_load: file for log F0 pdfs is not specified.\n");
   if (fp_ms_mcp == NULL)
      HTS_error(1,
                "HTS_HTS_ModelSet_load: file for mel-cepstrum pdfs is not specified.\n");

   /*-------------------- load pdfs for duration --------------------*/
   /* read the number of states & the number of pdfs (leaf nodes) */

   /* read the number of HMM states */
   HTS_fread_big_endian(&ms->nstate, sizeof(int), 1, fp_ms_dur);
   if (ms->nstate < 0)
      HTS_error(1,
                "HTS_HTS_ModelSet_load: #HMM states must be positive value.\n");

   /* read the number of duration pdfs */
   HTS_fread_big_endian(&ms->ndurpdf, sizeof(int), 1, fp_ms_dur);
   if (ms->ndurpdf < 0)
      HTS_error(1,
                "HTS_HTS_ModelSet_load: #duration pdf must be positive value.\n");

   ms->durpdf = (double **) HTS_calloc(ms->ndurpdf, sizeof(double *));
   ms->durpdf--;

   /* read pdfs (mean & variance) */
   for (i = 1; i <= ms->ndurpdf; i++) {
      ms->durpdf[i] = (double *) HTS_calloc(2 * ms->nstate, sizeof(double));
      ms->durpdf[i] -= 2;
      for (j = 2; j < 2 * (ms->nstate + 1); j++) {
         HTS_fread_big_endian(&temp, sizeof(float), 1, fp_ms_dur);
         ms->durpdf[i][j] = (double) temp;
      }
   }

   /*-------------------- load pdfs for mel-cepstrum --------------------*/
   /* read vector size for spectrum */
   HTS_fread_big_endian(&ms->mcpvsize, sizeof(int), 1, fp_ms_mcp);
   if (ms->mcpvsize < 0)
      HTS_error(1,
                "HTS_HTS_ModelSet_load: vector size of mel-cepstrum part must be positive.\n");

   ms->nmcppdf = (int *) HTS_calloc(ms->nstate, sizeof(int));
   ms->nmcppdf -= 2;

   /* read the number of pdfs for each state position */
   HTS_fread_big_endian(&ms->nmcppdf[2], sizeof(int), ms->nstate, fp_ms_mcp);
   for (i = 2; i <= ms->nstate + 1; i++) {
      if (ms->nmcppdf[i] < 0)
         HTS_error(1,
                   "HTS_HTS_ModelSet_load: #mel-cepstrum pdf at state %d must be positive value.\n",
                   i);
   }
   ms->mcppdf = (double ***) HTS_calloc(ms->nstate, sizeof(double **));
   ms->mcppdf -= 2;

   /* read pdfs (mean, variance) */
   for (i = 2; i <= ms->nstate + 1; i++) {
      ms->mcppdf[i] = (double **) HTS_calloc(ms->nmcppdf[i], sizeof(double *));
      ms->mcppdf[i]--;
      for (j = 1; j <= ms->nmcppdf[i]; j++) {
         ms->mcppdf[i][j] =
             (double *) HTS_calloc(2 * ms->mcpvsize, sizeof(double));
         for (k = 0; k < 2 * ms->mcpvsize; k++) {
            HTS_fread_big_endian(&temp, sizeof(float), 1, fp_ms_mcp);
            ms->mcppdf[i][j][k] = (double) temp;
         }
      }
   }

   /*-------------------- load pdfs for log F0 --------------------*/
   /* read the number of streams for f0 modeling */
   HTS_fread_big_endian(&ms->lf0stream, sizeof(int), 1, fp_ms_lf0);
   if (ms->lf0stream < 0)
      HTS_error(1,
                "HTS_HTS_ModelSet_load: #stream for log f0 part must be positive value.\n");

   ms->nlf0pdf = (int *) HTS_calloc(ms->nstate, sizeof(int));
   ms->nlf0pdf -= 2;

   /* read the number of pdfs for each state position */
   HTS_fread_big_endian(&ms->nlf0pdf[2], sizeof(int), ms->nstate, fp_ms_lf0);
   for (i = 2; i <= ms->nstate + 1; i++) {
      if (ms->nlf0pdf[i] < 0)
         HTS_error(1,
                   "HTS_HTS_ModelSet_load: #lf0 pdf at state %d must be positive.\n",
                   i);
   }
   ms->lf0pdf = (double ****) HTS_calloc(ms->nstate, sizeof(double ***));
   ms->lf0pdf -= 2;

   /* read pdfs (mean, variance & weight) */
   for (i = 2; i <= ms->nstate + 1; i++) {
      ms->lf0pdf[i] =
          (double ***) HTS_calloc(ms->nlf0pdf[i], sizeof(double **));
      ms->lf0pdf[i]--;
      for (j = 1; j <= ms->nlf0pdf[i]; j++) {
         ms->lf0pdf[i][j] =
             (double **) HTS_calloc(ms->lf0stream, sizeof(double *));
         ms->lf0pdf[i][j]--;
         for (k = 1; k <= ms->lf0stream; k++) {
            /* 4 -> mean, variance, voiced weight, and unvoiced weight */
            ms->lf0pdf[i][j][k] = (double *) HTS_calloc(4, sizeof(double));
            for (l = 0; l < 4; l++) {
               HTS_fread_big_endian(&temp, sizeof(float), 1, fp_ms_lf0);
               ms->lf0pdf[i][j][k][l] = (double) temp;
            }

            vw = ms->lf0pdf[i][j][k][2];        /* voiced weight */
            uvw = ms->lf0pdf[i][j][k][3];       /* unvoiced weight */
            if (vw < 0.0 || uvw < 0.0 || vw + uvw < 0.99 || vw + uvw > 1.01)
               HTS_error(1,
                         "HTS_HTS_ModelSet_load: voiced/unvoiced weights must be within 0.99 to 1.01.\n");
         }
      }
   }
}

/* HTS_HTS_ModelSet_clear: clear models */
void HTS_HTS_ModelSet_clear(HTS_ModelSet * ms)
{
   int i, j, k;

   /* free models for f0 */
   for (i = ms->nstate + 1; i >= 2; i--) {
      for (j = ms->nlf0pdf[i]; j >= 1; j--) {
         for (k = ms->lf0stream; k >= 1; k--) {
            HTS_free(ms->lf0pdf[i][j][k]);
         }
         HTS_free(++ms->lf0pdf[i][j]);
      }
      HTS_free(++ms->lf0pdf[i]);
      fflush(stdout);
   }
   ms->lf0pdf += 2;
   ms->nlf0pdf += 2;
   HTS_free(ms->lf0pdf);
   HTS_free(ms->nlf0pdf);

   /* free models for spectrum */
   for (i = ms->nstate + 1; i >= 2; i--) {
      for (j = ms->nmcppdf[i]; j >= 1; j--) {
         HTS_free(ms->mcppdf[i][j]);
      }
      HTS_free(++ms->mcppdf[i]);
   }
   ms->mcppdf += 2;
   ms->nmcppdf += 2;
   HTS_free(ms->mcppdf);
   HTS_free(ms->nmcppdf);

   /* free models for duration */
   for (i = ms->ndurpdf; i >= 1; i--) {
      ms->durpdf[i] += 2;
      HTS_free(ms->durpdf[i]);
   }
   HTS_free(++ms->durpdf);
}

/* HTS_Model_determine_lf0: determine lf0 parameter for final speaker */
void HTS_Model_determine_lf0(HTS_Model * m, HTS_ModelSet * ms,
                             HTS_TreeSet * ts, const int num_interp,
                             double uvthresh)
{
   int i, j, state, stream, lf0Index;
   const int nstate_plus1 = ms[0].nstate + 1;
   double mixweight;
   HTS_Tree *tree;

   for (state = 2; state <= nstate_plus1; state++) {
      for (stream = 1; stream <= ms[0].lf0stream; stream++) {
         m->lf0mean[state][stream] = m->lf0variance[state][stream] =
             mixweight = 0.0;
         for (i = 0; i < num_interp; i++) {
            tree = ts[i].thead[LF0];
            for (j = state; j > 2; j--) {
               tree = tree->next;
            }
            lf0Index = HTS_Node_search(tree->root, m->name);
            m->lf0mean[state][stream] +=
                ms[i].weight_interp * ms[i].lf0pdf[state][lf0Index][stream][0];
            m->lf0variance[state][stream] +=
                ms[i].weight_interp * ms[i].weight_interp *
                ms[i].lf0pdf[state][lf0Index][stream][1];
            mixweight +=
                ms[i].weight_interp * ms[i].lf0pdf[state][lf0Index][stream][2];
         }
         if (stream == 1) {
            if (mixweight > uvthresh)
               m->voiced[state] = TRUE;
            else
               m->voiced[state] = FALSE;
         }
      }
   }
}

/* HTS_Model_determine_mcp: determine mel-cepstrum parameter for final speaker (m) */
void HTS_Model_determine_mcp(HTS_Model * m, HTS_ModelSet * ms,
                             HTS_TreeSet * ts, const int num_interp)
{
   int i, j, state, dim, mcpIndex;
   int mvsize = ms[0].mcpvsize;
   HTS_Tree *tree;
   const int nstate_plus1 = ms[0].nstate + 1;

   for (state = 2; state <= nstate_plus1; state++) {
      m->mcpmean[state] = (double *) HTS_calloc(ms[0].mcpvsize, sizeof(double));
      m->mcpvariance[state] =
          (double *) HTS_calloc(ms[0].mcpvsize, sizeof(double));
      for (dim = 0; dim < ms[0].mcpvsize; dim++) {
         m->mcpmean[state][dim] = m->mcpvariance[state][dim] = 0.0;
         for (i = 0; i < num_interp; i++) {
            tree = ts[i].thead[MCP];
            for (j = state; j > 2; j--) {
               tree = tree->next;
            }
            mcpIndex = HTS_Node_search(tree->root, m->name);
            m->mcpmean[state][dim] +=
                ms[i].weight_interp * ms[i].mcppdf[state][mcpIndex][dim];
            m->mcpvariance[state][dim] +=
                ms[i].weight_interp * ms[i].weight_interp *
                ms[i].mcppdf[state][mcpIndex][dim + mvsize];
         }
      }
   }
}

/* HTS_Model_determine_dur: determine state duration from pdf */
void HTS_Model_determine_dur(HTS_Model * m, HTS_ModelSet * ms,
                             double *durmean, double *durvar,
                             double rho, double *diffdur)
{
   double data;
   int s;
   const int nstate_plus1 = ms->nstate + 1;

   for (s = 2; s <= nstate_plus1; s++) {
      data = durmean[s] + rho * durvar[s];

      m->dur[s] = (int) (data + *diffdur + 0.5);
      m->dur[s] = (m->dur[s] < 1) ? 1 : m->dur[s];

      m->totaldur += m->dur[s];
      *diffdur += data - (double) m->dur[s];
   }
}

/* HTS_Model_estimate_dur: determine duration parameter for final speaker */
void HTS_Model_estimate_dur(HTS_Model * m, HTS_ModelSet * ms,
                            HTS_TreeSet * ts, const int num_interp,
                            double *durmean, double *durvar)
{
   int i, state, durIndex;
   int nState = ms[0].nstate;

   for (state = 2; state <= nState + 1; state++) {
      durmean[state] = durvar[state] = 0.0;
      for (i = 0; i < num_interp; i++) {
         durIndex = HTS_Node_search(ts[i].thead[DUR]->root, m->name);
         durmean[state] += ms[i].weight_interp * ms[i].durpdf[durIndex][state];
         durvar[state] +=
             ms[i].weight_interp * ms[i].weight_interp *
             ms[i].durpdf[durIndex][state + nState];
      }
   }
}
