/*  ---------------------------------------------------------------  */
/*           The HMM-Based Speech Synthesis System (HTS)             */
/*                       HTS Working Group                           */
/*                                                                   */
/*                  Department of Computer Science                   */
/*                  Nagoya Institute of Technology                   */
/*                               and                                 */
/*   Interdisciplinary Graduate School of Science and Engineering    */
/*                  Tokyo Institute of Technology                    */
/*                                                                   */
/*                     Copyright (c) 2001-2007                       */
/*                       All Rights Reserved.                        */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The source code must retain the above copyright notice,     */
/*       this list of conditions and the following disclaimer.       */
/*                                                                   */
/*    2. Any modifications to the source code must be clearly        */
/*       marked as such.                                             */
/*                                                                   */
/*    3. Redistributions in binary form must reproduce the above     */
/*       copyright notice, this list of conditions and the           */
/*       following disclaimer in the documentation and/or other      */
/*       materials provided with the distribution.  Otherwise, one   */
/*       must contact the HTS working group.                         */
/*                                                                   */
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF TECHNOLOGY,   */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF         */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */

/* $Id: HTS_tree.c,v 1.2 2007/09/28 05:40:09 zen Exp $ */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* hts_engine libraries */
#include "HTS_hidden.h"

/* HTS_dp_match: recursive matching */
static HTS_Boolean HTS_dp_match(char *str, char *pat, const int pos,
                                const int max)
{
   if (pos > max)
      return FALSE;
#ifdef FAST_TREE_MATCH
   if (str[0] == '\0' && pat[0] == '\0')
#else
   if (*str == '\0' && *pat == '\0')
#endif                          /* FAST_TREE_MATCH */
      return TRUE;

#ifdef FAST_TREE_MATCH
   if (pat[0] == '*') {
#else
   if (*pat == '*') {
#endif                          /* FAST_TREE_MATCH */
      if (HTS_dp_match(str + 1, pat, pos + 1, max) == 1)
         return TRUE;
      else
         return HTS_dp_match(str + 1, pat + 1, pos + 1, max);
   }
#ifdef FAST_TREE_MATCH
   if (str[0] == pat[0] || pat[0] == '?') {
#else
   if (*str == *pat || *pat == '?') {
#endif                          /* FAST_TREE_MATCH */
      if (HTS_dp_match(str + 1, pat + 1, pos + 1, max + 1) == 1)
         return TRUE;
      else {
#ifdef FAST_TREE_MATCH
         if (pat[1] == '*')
#else
         if (*(pat + 1) == '*')
#endif                          /* FAST_TREE_MATCH */
            return HTS_dp_match(str + 1, pat + 2, pos + 1, max + 1);
      }
   }

   return FALSE;
}

/* HTS_pattern_match: pattern matching function */
static HTS_Boolean HTS_pattern_match(char *str, char *pat)
{
   int i, max = 0, nstar = 0, nques = 0;
   char buf[HTS_MAXBUFLEN];

#ifdef FAST_TREE_MATCH
   const int pat_len = strlen(pat);
   int j;
   int buf_len;
   for (i = 0; i < pat_len; i++) {
#else
   for (i = 0; i < (int) strlen(pat); i++) {
#endif                          /* FAST_TREE_MATCH */
      switch (pat[i]) {
      case '*':
         nstar++;
         break;
      case '?':
         nques++;
         max++;
         break;
      default:
         max++;
      }
   }

   if (nstar == 2 && nques == 0 && pat[0] == '*' && pat[i - 1] == '*') {
      /* only string matching is required */
#ifdef FAST_TREE_MATCH
      buf_len = i - 2;
      for (i = 0, j = 1; i < buf_len; i++, j++)
         buf[i] = pat[j];
      buf[buf_len] = '\0';
#else
      strncpy(buf, &pat[1], i - 2);
      buf[i - 2] = '\0';
#endif                          /* FAST_TREE_MATCH */
      if (strstr(str, buf) != NULL)
         return TRUE;
      else
         return FALSE;
   } else
      return HTS_dp_match(str, pat, 0, (int) (strlen(str) - max));
}

/* HTS_Question_match: check given string match given question */
static HTS_Boolean HTS_Question_match(HTS_Question * q, char *str)
{
   HTS_Boolean flag = FALSE;
   HTS_Pattern *p;

   for (p = q->phead;; p = p->next) {
      flag = HTS_pattern_match(str, p->pat);
      if (flag)
         return TRUE;
      if (p == q->ptail)
         break;
   }

   return FALSE;
}

/* HTS_Node_search: tree search */
int HTS_Node_search(HTS_Node * node, char *str)
{
   while (node != NULL) {
      if (HTS_Question_match(node->quest, str)) {
         if (node->yes->pdf > 0)
            return (node->yes->pdf);
         node = node->yes;
      } else {
         if (node->no->pdf > 0)
            return (node->no->pdf);
         node = node->no;
      }
   }
   return -1;
}

/* HTS_Question_load: Load questions from file */
static void HTS_Question_load(HTS_Question * q, FILE * fp, const HTS_Mtype type)
{
   char buf[HTS_MAXBUFLEN];
   HTS_Pattern *p;

   HTS_get_token(fp, buf);
   q->qName = HTS_strdup(buf);
   q->phead = NULL;

   HTS_get_token(fp, buf);
   if (strcmp(buf, "{") == 0) {
      while (1) {
         HTS_get_token(fp, buf);
         p = (HTS_Pattern *) HTS_calloc(1, sizeof(HTS_Pattern));
         if (q->phead) {
            q->ptail->next = p;
            q->ptail = p;
         } else {
            q->phead = q->ptail = p;
         }
         q->ptail->pat = HTS_strdup(buf);
         HTS_get_token(fp, buf);
         if (!strcmp(buf, "}"))
            break;
      }
   }
}

/* HTS_Tree_parse_pat: parse pattern specified for each tree */
static void HTS_Tree_parse_pat(HTS_Tree * tree, char *buf)
{
   char *s, *l, *r;
   HTS_Pattern *p;

   tree->phead = NULL;

   /* parse tree pattern */
   s = buf;
   if ((l = strchr(s, '{')) != NULL) {  /* pattern is specified */
      s = l + 1;
      if (*s == '(')
         ++s;

      r = strrchr(s, '}');
      if (s < r && *(r - 1) == ')')
         --r;
      *r = ',';

      /* parse pattern */
      while ((l = strchr(s, ',')) != NULL) {
         p = (HTS_Pattern *) HTS_calloc(1, sizeof(HTS_Pattern));
         if (tree->phead) {
            tree->ptail->next = p;
            tree->ptail = p;
         } else {
            tree->phead = tree->ptail = p;
         }
         *l = '\0';
         tree->ptail->pat = HTS_strdup(s);
         s = l + 1;
      }
   }
}

/* HTS_get_state_num: return the number of state */
static int HTS_get_state_num(char *buf)
{
   char *s, *l, *r;
   int state;

   s = buf;
   if (((l = strchr(s, '[')) == NULL)
       || ((r = strrchr(s, ']')) == NULL)) {
      return 0;
   } else {
      *r = '\0';
      s = l + 1;
      state = atoi(s);
   }

   return state;
}

/* HTS_is_num: check given buffer is number or not */
static HTS_Boolean HTS_is_num(char *buf)
{
   int i;

   for (i = 0; i < (int) strlen(buf); i++)
      if (!(isdigit(buf[i]) || (buf[i] == '-')))
         return FALSE;

   return TRUE;
}

/* HTS_TreeSet_find_question: find question from question list */
static HTS_Question *HTS_TreeSet_find_question(HTS_TreeSet * ts,
                                               const HTS_Mtype type, char *buf)
{
   HTS_Question *q;

   for (q = ts->qhead[type];; q = q->next) {
      if (strcmp(buf, q->qName) == 0)
         break;
      if (q == ts->qtail[type]) {
         q = NULL;
         break;
      }
   }

   if (!q)
      HTS_error(1,
                "HTS_TreeSet_find_question: cannot find question %s.\n", buf);

   return q;
}

/* HTS_name2num: convert name of node to number */
static int HTS_name2num(char *buf)
{
   return (atoi(strrchr(buf, '_') + 1));
}

/* HTS_Node_find: find node for given number */
static HTS_Node *HTS_Node_find(HTS_Node * node, const int num)
{
   while (node) {
      if (node->idx == num)
         return node;
      else
         node = node->next;
   }
   return NULL;
}

/* HTS_Tree_load: Load trees */
static void HTS_Tree_load(HTS_Tree * tree, FILE * fp, HTS_TreeSet * ts,
                          const HTS_Mtype type)
{
   char buf[HTS_MAXBUFLEN];
   HTS_Node *node;

   HTS_get_token(fp, buf);
   node = (HTS_Node *) HTS_calloc(1, sizeof(HTS_Node));
   tree->root = tree->leaf = node;

   if (strcmp(buf, "{") == 0) {
      while (HTS_get_token(fp, buf), strcmp(buf, "}") != 0) {
         node = HTS_Node_find(tree->leaf, atoi(buf));
         HTS_get_token(fp, buf);        /* load question at this node */

         node->quest = HTS_TreeSet_find_question(ts, type, buf);
         node->yes = (HTS_Node *) HTS_calloc(1, sizeof(HTS_Node));
         node->no = (HTS_Node *) HTS_calloc(1, sizeof(HTS_Node));

         HTS_get_token(fp, buf);
         if (HTS_is_num(buf)) {
            node->no->idx = atoi(buf);
         } else {
            node->no->pdf = HTS_name2num(buf);
         }
         node->no->next = tree->leaf;
         tree->leaf = node->no;

         HTS_get_token(fp, buf);
         if (HTS_is_num(buf)) {
            node->yes->idx = atoi(buf);
         } else {
            node->yes->pdf = HTS_name2num(buf);
         }
         node->yes->next = tree->leaf;
         tree->leaf = node->yes;
      }
   } else {
      node->pdf = HTS_name2num(buf);
   }
}

/* HTS_TreeSet_load: Load decision tree information */
void HTS_TreeSet_load(HTS_TreeSet * ts, FILE * fp, const HTS_Mtype type)
{
   char buf[HTS_MAXBUFLEN];
   HTS_Question *q;
   HTS_Tree *t;
   int state;

   /* check */
   if (fp == NULL)
      HTS_error(1, "HTS_TreeSet_load: file for trees is not specified.\n");

   ts->nTrees[type] = 0;
   ts->qhead[type] = NULL;
   ts->thead[type] = NULL;
   /* parse tree files */
   while (!feof(fp)) {
      HTS_get_token(fp, buf);
      /* parse questions */
      if (strcmp(buf, "QS") == 0) {
         q = (HTS_Question *) HTS_calloc(1, sizeof(HTS_Question));
         HTS_Question_load(q, fp, type);
         if (ts->qhead[type]) {
            ts->qtail[type]->next = q;
            ts->qtail[type] = q;
         } else {
            ts->qhead[type] = ts->qtail[type] = q;
         }
         q->next = NULL;
      }
      /* parse trees */
      if ((state = HTS_get_state_num(buf)) != 0) {
         t = (HTS_Tree *) HTS_calloc(1, sizeof(HTS_Tree));
         t->state = state;
         t->next = NULL;
         t->root = t->leaf = NULL;
         t->phead = t->ptail = NULL;
         t->state = 0;
         HTS_Tree_parse_pat(t, buf);
         HTS_Tree_load(t, fp, ts, type);
         if (ts->thead[type]) {
            ts->ttail[type]->next = t;
            ts->ttail[type] = t;
         } else {
            ts->thead[type] = ts->ttail[type] = t;
         }
         t->next = NULL;
         /* increment # of trees */
         ts->nTrees[type]++;
      }
   }

   /* no tree information in tree file */
   if (ts->thead[type] == NULL) {
      if (type == DUR)
         HTS_error(1, "HTS_TreeSet_load: no trees for duration are loaded.\n");
      else if (type == LF0)
         HTS_error(1, "HTS_TreeSet_load: no trees for log f0 are loaded.\n");
      else
         HTS_error(1,
                   "HTS_TreeSet_load: no trees for mel-cepstrum are loaded.\n");
   }
}

/* HTS_Node_clear: recursive function to free Node */
static void HTS_Node_clear(HTS_Node * node)
{
   if (node->yes != NULL)
      HTS_Node_clear(node->yes);
   if (node->no != NULL)
      HTS_Node_clear(node->no);

   HTS_free(node);
}

/* HTS_Tree_clear: clear given tree */
static void HTS_Tree_clear(HTS_Tree * tree)
{
   HTS_Pattern *phead, *p;

   phead = tree->phead;
   if (phead != NULL) {
      while (phead != tree->ptail) {
         p = phead;
         phead = p->next;
         HTS_free(p->pat);
         HTS_free(p);
      }
      if (phead->pat != NULL)
         HTS_free(phead->pat);
      HTS_free(phead);
   }

   HTS_Node_clear(tree->root);
}

/* HTS_Question_clear: clear loaded question */
static void HTS_Question_clear(HTS_Question * q)
{
   HTS_Pattern *p, *next = NULL;

   for (p = q->phead;; p = next) {
      HTS_free(p->pat);
      if (p == q->ptail) {
         HTS_free(p);
         break;
      } else {
         next = p->next;
         HTS_free(p);
      }
   }
   HTS_free(q->qName);
}

/* HTS_TreeSet_clear: clear decision trees */
void HTS_TreeSet_clear(HTS_TreeSet * ts, const HTS_Mtype type)
{
   HTS_Question *q, *qnext = NULL;
   HTS_Tree *t, *tnext = NULL;

   /* free questions */
   for (q = ts->qhead[type];; q = qnext) {
      HTS_Question_clear(q);
      if (q == ts->qtail[type]) {
         HTS_free(q);
         break;
      } else {
         qnext = q->next;
         HTS_free(q);
      }
   }

   /* free trees */
   for (t = ts->thead[type];; t = tnext) {
      HTS_Tree_clear(t);
      if (t == ts->ttail[type]) {
         HTS_free(t);
         break;
      } else {
         tnext = t->next;
         HTS_free(t);
      }
   }
}
